#include <stdio.h>			/* **/
#include <stdarg.h>			/* **/
#include <string.h>			/* **/
#include <malloc.h>			/* **/
#include <stdlib.h>			/* **/
#include <unistd.h>			/* **/
#include <time.h>			/* **/
#include <errno.h>			/* **/
#include <ctype.h>			/* **/
#include <fcntl.h>			/* **/
#include <pwd.h>			/* **/	
#include <termios.h>			/* **/
#include <signal.h>			/* **/
#include <sys/errno.h>			/* **/
#include <sys/time.h>			/* **/
#include <sys/types.h>			/* **/
#include "ustcio.hpp"			/* ustc io functions **/
#include "ustclib.hpp"			/* ustc cad functions **/
#include "ustcpdsn.hpp"			/* ustc padsnet functions **/
#define USTC_USAGE \
"USTC %d.%02d (c) 2006-2010 Dmitriy I.Cherkashin, www.ucrouter.ru\n"  \
"email: dch@ucrouter.ru , divch@users.sourceforge.net , dch@ucrouter.com\n"  \
"phone: +8-909-936-8809; ICQ: 474-363-900;\n"  \
"This is not public software, you can't use this program without\n"  \
"license agreement, which available via option --license\n"  \
"with noted above author. To see license agreement use:\n"  \
"ustc --license\n"  \
"  or\n"  \
"ustc --license-ru\n"  \
"  for russian license translation.\n"
#define USTC_LICENSE \
"USTC %d.%02d (c) 2006-2010 Dmitriy I.Cherkashin, www.ucrouter.ru\r\n"  \
"email: dch@ucrouter.ru , divch@users.sourceforge.net , dch@ucrouter.com\r\n"  \
"phone: +8-909-936-8809; ICQ: 474-363-900;\r\n"  \
"This program is free software with source codes\r\n"  \
"downloadable via internation communicational net.\r\n"  \
"If you have source codes of this software original or modified,\r\n"  \
"you must provide ability to free receive original source codes\r\n"  \
"of this software to another users even in internation communicational\r\n"  \
"net absense without any unusual additional payment exept\r\n" \
"downloading files from internation communicational net cost or\r\n" \
"mail cost.\r\n"
#define USTC_LICENSE_RU \
"USTC %d.%02d (c) 2006-2010  ., www.ucrouter.ru\r\n" \
" : dch@ucrouter.ru,divch@users.sourceforge.net,dch@ucrouter.com\r\n"  \
": +8-909-936-8809; ICQ: 474-363-900;\r\n" \
"     \r\n" \
"      \r\n" \
" .\r\n" \
"       \r\n" \
"  ,     \r\n" \
"     \r\n" \
"     \r\n" \
"      \r\n" \
"       \r\n" \
"    \r\n" \
"     .\r\n"
#define  __USTC_DEBUG__
#ifdef  __USTC_DEBUG__
u_int8_t ustc_debug_mode = 0;
#endif 							/* __USTC_DEBUG__ **/
u_int8_t ustc_version_hi = 1; 		/* version hight number **/
u_int8_t ustc_version_lo = 40;		/* version low **/
u_int8_t ustc_nl_print = 0;
u_int8_t ustc_local = 0;		/* take executable files from local **/
FILE * ustc_fi = 0;			/* input file descriptor **/
FILE * ustc_fo = 0;			/* output file descriptor **/
#define USTC_FILE_NAME_SIZE	1024
u_int8_t ustc_part[256] = {0};		/* part name */
char ustc_ifile[USTC_FILE_NAME_SIZE] = ""; /* input file name **/
char ustc_ofile[USTC_FILE_NAME_SIZE] = ""; /* output file name **/
/* file format **/
u_int8_t ustc_ifmt = 1; 		/* input file format **/
u_int8_t ustc_ofmt = 1; 		/* output file format **/
#define USTC_UNKNOWN  	0		/* unknown format **/
#define USTC_LIB      	5		/* lib file format **/
#define USTC_PADS_NET 	7		/* pads net file format */
u_int8_t * ustc_ibuf = 0;	/* input buffer **/
u_int8_t * ustc_obuf = 0;	/* output buffer **/
u_int32_t ustc_isize = 0;	/* input buffer size **/
u_int32_t ustc_osize = 0;	/* output buffer size **/
u_int8_t ustc_sort_enable = 1;		/* output file sort enable */
/* error message **/
#if 1
int ustc_error(const char *frmt, ...) 
{
 char pbuf[512];			/* message buffer **/
 va_list valt;				/* argument list **/
 va_start(valt, frmt);			/* init argument list **/
 vsnprintf(pbuf, sizeof(pbuf), frmt, valt);	/* **/
 va_end(valt);				/* **/
 fprintf(stderr, "%s", pbuf);	/* print error message to stderr **/	
 fflush(stderr);			
 return(1);
}
#else
#define ustc_error(...) 
#endif							/* **/
int ustc_debug(const char *frmt, ...) 	/* print debug message & exit **/
{
 char pbuf[1024];			/* message buffer **/
/* from command args **/
 va_list valt;				/* argument list **/
 va_start(valt, frmt);			/* init argument list **/
 vsnprintf(pbuf, sizeof(pbuf), frmt, valt);	/* **/
 va_end(valt);				/* **/
 fprintf(stderr, "%s", pbuf);	/* print error message to stderr **/	
 fflush(stderr);			
 return(1);
}
#if 1
int ustc_printf( const char *frmt, ... ) 
{
 static char pbuf[4096];		/* print buffer **/
 va_list valt;				/* argument list **/
 va_start(valt, frmt);			/* init argument list **/
 vsnprintf(pbuf, sizeof(pbuf), frmt, valt);
 va_end(valt);
 fprintf(stdout, "%s", pbuf);
 fflush(stdout);
 return(1);
}
#else /* 0 **/
#define ustc_printf printf
#endif							
int ustc_usage(void)			/* print help message **/
{
 ustc_printf(
    USTC_USAGE
    "Usage : ustc [options] <input file> [<output file>]\r\n"
    "where options may be:\r\n"
    "--input-format lib, lib - kicad schematic library file\r\n"
    "--output-format lib, lib - kicad schematic library file\r\n"
    "--sort on | off - output report sort enable or disable\r\n"
    "--part <part name> - create & modify kicad sch part\r\n"
    "--pin-name-to-text <part name> - convert kicad sch part pinm names to text\r\n"
    "--debug        - print debug information\r\n"
    "--version      - print version and exit\r\n"
    "--todo         - print todo and exit\r\n"
    "--changelog    - print changelog and exit\r\n"
    "--man          - print man page and exit\r\n"
    "--license      - print program usage license\r\n"
    "--license-ru   - print program usage license, russian translation\r\n"
    "--help         - print help note and exit\r\n"
    "Files extensions net - PADS-PCB netlist, lib - Kicad component netlist"
    ,
	ustc_version_hi,		/* version hight number **/
	ustc_version_lo			/* version low **/
 );
 return(1);
}
int ustc_version(void)	/* print version **/
{
 ustc_printf(
    "\r\nVersion %d.%02d\r\n",
    ustc_version_hi,			/* version hight number **/
    ustc_version_lo			/* version low **/
 );					
 return(1);
}
int ustc_todo(void)		/* print todo note **/
{
 char * msg = 
    "";
 ustc_printf(msg);
 return(1);
}
int ustc_changelog(void)		/* print changelog note **/
{
 ustc_printf(
    "\r\nChangeLog\r\n"
    "- 2010\06\05 PAD NET format added\r\n"
 );
 return(1);
}
int ustc_man(void)			/* **/
{
 ustc_printf(
    "\nman\n"
    "\n"
 );
 return(1);
}
int ustc_license(void)
{
/* cutted version - if you have source codes of this software,
you must provide this ability to another users **/
 ustc_printf(USTC_LICENSE,
    ustc_version_hi,			/* version hight number **/
    ustc_version_lo			/* version low **/
 );
 return(1);
}
int ustc_license_ru(void)
{
/*   -        ,
        **/
 ustc_printf(USTC_LICENSE_RU,
    ustc_version_hi,			/* version hight number **/
    ustc_version_lo			/* version low **/
 );
 return(1);
}
int ustc_format(char * nam)		/* return file format **/
{
 if(nam == 0) {
    return(USTC_UNKNOWN);		/* unknown format **/
 }
 char *bnam;				/* base name **/
 bnam = basename(nam);			/* read base name **/
 if(bnam == 0) {
    return(USTC_UNKNOWN);		/* unknown format **/
 }
 while(*bnam != '\0') {
    if(*bnam == '.') {
    /* file extension **/
	bnam += 1;			/* next character **/
	if(
	    (strcmp(bnam,"LIB") == 0) ||
	    (strcmp(bnam,"lib") == 0)
	) {
	    return(USTC_LIB);
	}
	else {
	    return(USTC_UNKNOWN);	/* unknown format **/
	}
    }
    bnam += 1;
 }
 return(USTC_UNKNOWN);			/* unknown format **/
}
static char * ustc_sig_name(siginfo_t * sinfo, char * sname, int ssize) /* print signal name **/
{
 static struct {
    int num;				/* signal number **/
    char * name;			/* signal name **/
 } sginfo[] = {
    { SIGHUP   , "1)SIGHUP - Hangup" },
    { SIGINT   , "2)SIGINT - Interrupt" },
    { SIGQUIT  , "3)SIGQUIT - Quit" },
    { SIGILL   , "4)SIGILL - Illegal instruction" },
    { SIGTRAP  , "5)SIGTRAP - trap" },
    { SIGABRT  , "6)SIGABRT - abort" },
    { SIGBUS   , "7)SIGBUS - buss error" },
    { SIGFPE   , "8)SIGFPE - floating point exeption" },
/* not processed signal **/ 
    { SIGKILL  , "9)SIGKILL - kill" },
    { SIGUSR1  , "10)SIGUSR1 - user signal 1" },
    { SIGSEGV  , "11)SIGSEGV - segmetation violation" },
    { SIGUSR2  , "12)SIGUSR2 - user signal 2" },
    { SIGPIPE  , "13)SIGPIPE - pipe" },
    { SIGALRM  , "14)SIGALRM - alarm" },
    { SIGTERM  , "15)SIGTERM - term" },
    { SIGSTKFLT, "16)SIGSTKFLT - stack failt" },
/* 17-31 **/
    { SIGCHLD  , "17)SIGCHLD - children" },
    { SIGCONT  , "18)SIGCONT - continue" },
    { SIGSTOP  , "19)SIGSTOP - stop" },
    { SIGTSTP  , "20)SIGTSTP - terminal stop, terminal command" },
    { SIGTTIN  , "21)SIGTTIN - terminal read" },
    { SIGTTOU  , "22)SIGTTOU - terminal weite" },
    { SIGURG   , "23)SIGURG - urg" },	/*  **/
    { SIGXCPU  , "24)SIGXCPU - cpu limit exceeded" },
/* **/
    { SIGXFSZ  , "25)SIGXFSZ - file size exceeded" },
    { SIGVTALRM, "26)SIGVTALRM - virtual alarm clock" },
    { SIGPROF  , "27)SIGPROF - profilling alarm clock" },
    { SIGWINCH , "28)SIGWINCH - window size changed" },
    { SIGIO    , "29)SIGIO - I/O" },
    { SIGPWR   , "30)SIGPWR - power failure restart" },
    { SIGSYS   , "31)SIGSYS - bad system call" }
 };
 char * sptr;				/* signal name pointer **/
 if(sinfo->si_signo >= 1 && sinfo->si_signo <= 31) {
    if(sginfo[sinfo->si_signo-1].num == sinfo->si_signo) { /* print signal description **/
	sptr = sginfo[sinfo->si_signo-1].name;
    } else {
	return(0);
    }
 } else {
    return(0);
 }
 if(sname == 0 || ssize <= 1) {
    return(sptr);
 }
 snprintf(
    sname,				/* signal name buffer **/
    ssize,				/* buffer size **/
    "%s",				/* format **/
    sptr
 );    
 return(sname);				/* return buffer pointer **/
} 
static int ustc_sig_info(siginfo_t * sinfo, void * sptr) /* print signal information **/
{
 if((sinfo == 0) || (sptr == 0)) {
/* null pointers -> print error message and exit **/
    ustc_error("Bad signal call sinfo=0x%x , sptr=0x%x\n",
	sinfo	,		/* signal pointer **/
	sptr			/* **/
    );
    return(0);
 }

 char sname[32];
 if(ustc_sig_name(sinfo,sname,sizeof(sname)) != 0) {
/* print signal description **/
    ustc_printf("%s\n", sname);
 }
/* print signal description **/
 ustc_printf(
    "Signal num {%d}, errno {%d}, code {%d}\n"
    "pid {%d}, uid {%d}, status {%d}, utime {%d}, stime {%d}\n"
    "value {%d}, int {%d}, ptr {%d}, addr {%d}, band {%d}, fd {%d}\n"
    "sptr {%x}\n",
/* **/
    sinfo->si_signo	,	/* signal number **/
    sinfo->si_errno	,	/* errno code **/
    sinfo->si_code 	,	/* code **/
/* **/
    sinfo->si_pid 	,	/* process id **/
    sinfo->si_uid	,	/* user id **/
    sinfo->si_status	,	/* **/
    sinfo->si_utime	,	/* user time **/
    sinfo->si_stime	,	/* system time **/
/* **/
    sinfo->si_value	,	/* signal vale **/
    sinfo->si_int 	,	/* int **/
    sinfo->si_ptr	,	/* ptr **/
    sinfo->si_addr	,	/* addr **/
    sinfo->si_band	,	/* **/
    sinfo->si_fd	,	/* file descriptor **/
    sptr			/* **/
 );
 return(1);	
}
int ustc_sig_log(siginfo_t * sinfo, void * sptr) /* log signal **/
{
 if(sinfo == 0) {			/* signal description null pointer **/	
    return(0);
 } 

 char * lname = "/var/log/ustcdump";	/* log file name **/
 FILE * flog;				/* log file **/
/* **/
 int rvl;				/* return code **/
/* **/
 rvl = 1;				/* start return code **/

 flog = fopen(lname, "r+b");		/* open log file  **/
 if(flog == 0) {
    flog = fopen(lname, "w");		/* open log file  **/
    if(flog == 0) {			/* open file in write mode error **/
	return(0);
    }
 } else {
/* file opened in read mode -> set to file end **/
    if(fseek(flog, 0, SEEK_END) < 0) {
	rvl = 0;
    }
 }
 char buf[32];    			/* signal name buffer **/
 if(ustc_sig_name(sinfo,buf,sizeof(buf)) <= 0) {
    fprintf(flog, "%d\n", sinfo->si_signo);
 } else {
    fprintf(flog, "%s\n", buf);
 }
    
 if(fclose(flog) < 0) {
    rvl = 0;
 }

 if(rvl == 0) {
    return(0);
 }
 return(1);
}
/* terminate signal received **/
volatile int kill_link = 0;
static void ustc_sig_term_ext(int sig, siginfo_t * sinfo, void * sptr)
{
 if(kill_link == 0) {
    kill_link = 1;			/* received signal **/
    ustc_sig_info(sinfo,sptr);		/* print signal description **/
    ustc_sig_log(sinfo,sptr);		/* log signal **/
 } else {
    kill_link += 1;			/* received signal **/
 }
}
volatile u_int8_t debug_alrm = 1;	/* print debug note in alarm handler **/
volatile u_int8_t 
sig_alrm = 0;		/* > 1, exit from select loop **/
static void ustc_sig_alarm_ext(int sig, siginfo_t * sinfo, void * sptr)
{
 if(sig_alrm == 0) {
    if(debug_alrm > 0) {
/* debug alarm signals **/
	ustc_sig_info(sinfo,sptr);	/* print signal description **/
	ustc_sig_log(sinfo,sptr);	/* log signal **/
    }
 }
 sig_alrm += 1;			/* > 1, exit from select loop **/
}
/* sigpipe processing function  **/
volatile int ustc_sig_pipe_ex = 0;
static void ustc_sig_pipe_ext(int sig, siginfo_t * sinfo, void * sptr)
{
 ustc_sig_info(sinfo,sptr);		/* print signal description **/
 ustc_sig_log(sinfo,sptr);		/* log signal **/
 ustc_sig_pipe_ex = 1;
}
/* SIGTERM/SIGINT **/
volatile int sig_chld_ex = 0;
static void ustc_sig_chld_ext(int sig, siginfo_t * sinfo, void * sptr)
{
 ustc_sig_info(sinfo,sptr);		/* print signal description **/
 ustc_sig_log(sinfo,sptr);		/* log signal **/
 sig_chld_ex = 1;
}
/* bad signal received **/
static void ustc_bad_signal_ext(int sig, siginfo_t * sinfo, void * sptr)
{
 ustc_sig_info(sinfo,sptr);		/* print signal description **/
 ustc_sig_log(sinfo,sptr);		/* log signal **/
 exit(1);		
}
int ustc_signal(int snum, void (*actn)(int, siginfo_t *, void *)) 
{
 if(
    (snum == SIGKILL) || 		/* unconditionall kill **/
    (snum == SIGSTOP)			/* stop process **/
 ) {
/* not processed signals **/
    return(0);
 } 
/* signal action structure **/
 struct sigaction sact;			/* signal action **/
 sact.sa_handler = 0;			/* signal handler **/
 sact.sa_sigaction = actn;		/* signal extended handler **/
 sact.sa_restorer = 0;			/* restore function **/
/* signal process **/
 sact.sa_flags = (
    SA_SIGINFO | 			/* signal handler requiry 3 arguments **/
    SA_RESTART | 			/* BSD compatible **/
    SA_NOMASK				/* not mask this signal **/
 );
/* clear blocked signal mask **/
 if(sigemptyset(&sact.sa_mask) < 0) {	/* clear mask error **/
    ustc_error("Clear nonblocked signals mask error\n");
    return(0);
 }
/* blocked signals mask **/ 
 if(
    (sigaddset(&sact.sa_mask, SIGHUP)  < 0) || 	/* SIGHUP **/
    (sigaddset(&sact.sa_mask, SIGINT)  < 0) ||	/* INT **/
    (sigaddset(&sact.sa_mask, SIGTERM) < 0) ||	/* TERM **/
    (sigaddset(&sact.sa_mask, SIGCHLD) < 0) ||	/* CHLD **/
    (sigaddset(&sact.sa_mask, SIGUSR2) < 0)   	/* USR2 **/
 ) {
/* add signal in mask error **/
    ustc_error("Set nonblocked signals mask %x error\n", sact.sa_mask);
    return(0);
 }
/* set signal handler **/
 if(sigaction(snum, &sact, 0) < 0) {
    ustc_error("Set signal %d handler error\n");
    return(0);
 }
 return(1);
}
USTC_LAYER * ustc_layer = 0;		/* layer buffer **/
u_int32_t ustc_layer_count = 0;		/* layer count **/
int ustc_str_sort(u_int8_t **nname, u_int32_t ncount)
{
 if(nname == 0) {			/* name buffer pointer check */
    return(0);
 }
 u_int32_t ntnum;			/* name number */
 ntnum = 0; 				/* name number start */
 while(ntnum < ncount) {		/* name loop */
    u_int32_t scnum;			/* second number */
    scnum = ntnum + 1; 			/* second number start */
    while(scnum < ncount) {		/* name second loop */
	int cmp;			/* compare result */
	if(*(nname + ntnum) == 0) {	/* name first pointer check */
	    cmp = (-1);			/* (<) */
	} else {
	    if(*(nname + scnum) == 0) { /* name second pointer check */
		cmp = 1;		/* (>) */
	    } else {
		cmp = strcmp(		/* string compare */
		    (char *)(*(nname + ntnum)), /* net first pointer */
		    (char *)(*(nname + scnum)) /* net second pointer */
		);
	    }
	} 
	if(cmp > 0) {			/* compare result test */
	    u_int8_t *tmname;
	    tmname = *(nname + ntnum);
	    *(nname + ntnum) = *(nname + scnum);
	    *(nname + scnum) = tmname;
	}
	scnum += 1; 			/* number second next */
    }    
    ntnum += 1; 			/* number first next */
 }
 return(1);
}
USTC_LIB_HEADER ustc_lib_header;
USTC_LIB_ALIAS * ustc_lib_alias = 0;	/* kicad lib alias pointer */
u_int32_t ustc_lib_alias_count = 0;	/* kicad lib alias count */
USTC_LIB_FIELD * ustc_lib_field = 0;	/* kicad lib field buffer pointer */
u_int32_t ustc_lib_field_count = 0;	/* kicad lib field count */
USTC_LIB_POLY * ustc_lib_poly = 0;	/* kicad lib point buffer */
u_int32_t ustc_lib_poly_count = 0;	/* kicad lib point count */
USTC_LIB_RECT * ustc_lib_rect = 0;	/* kicad lib rect buffer */
u_int32_t ustc_lib_rect_count = 0;	/* kicad lib rect count */
USTC_LIB_RING * ustc_lib_ring = 0;	/* kicad lib ring buffer pointer */
u_int32_t ustc_lib_ring_count = 0;	/* kicad lib ring count */
USTC_LIB_ARC * ustc_lib_arc = 0;	/* kicad lib arc buffer pointer */
u_int32_t ustc_lib_arc_count = 0;	/* kicad lib arc count */
USTC_LIB_TEXT * ustc_lib_text = 0;	/* kicad lib text buffer */
u_int32_t ustc_lib_text_count = 0;	/* kicad lib text count */
USTC_LIB_DRAW * ustc_lib_draw = 0;	/* kicad draw buffer */
u_int32_t ustc_lib_draw_count = 0;	/* kicad draw count */
USTC_LIB_PIN * ustc_lib_pin = 0;	/* cad pin buffer pointer */
u_int32_t ustc_lib_pin_count = 0;	/* cad pin count */
USTC_LIB_COMP * ustc_lib_comp = 0;	/* kicad lib component buffer */
u_int32_t ustc_lib_comp_count = 0;	/* kicad lib component count */
int ustc_lib_read(void)			/* lib read */
{
 int rvl; 				/* return code */
 rvl = 1;				/* return code start value */
 if(ustc_lib_header_clear(&ustc_lib_header) == 0) { /* lib header clear */
    if(rvl == 1) {			/* return code check */
	ustc_error(			/* error message print */
	    "Kicac lib header clear error\r\n"
	);
	rvl = 0;			/* return code clear */
    }
 }
 u_int8_t buf[256];			/* buffer */
 u_int32_t len;				/* buffer size */
 len = sizeof(buf);			/* buffer size read */
 if(ustc_lib_header_read(&ustc_lib_header, buf, len) == 0) {
    if(rvl == 1) {
	ustc_error("Kicac lib header read error\r\n");
	rvl = 0;
    }
 }
 u_int32_t cnum;			/* component number */
 u_int32_t alnum;			/* alias number */
 u_int32_t flnum;			/* field number */
 u_int32_t drnum;			/* draw number */
 u_int32_t plnum;			/* poly number */
 u_int32_t rcnum;			/* rectangle number */
 u_int32_t rngnum;			/* ring number */
 u_int32_t arnum;			/* arc number */
 u_int32_t txnum;			/* text number */
 u_int32_t pnnum;			/* pin number */
 alnum = ustc_lib_alias_count;		/* kicad lib alias count */
 flnum = ustc_lib_field_count;		/* kicad lib field count */
 plnum = ustc_lib_poly_count;		/* kicad lib point count */
 rcnum = ustc_lib_rect_count;		/* kicad lib rect count */
 rngnum = ustc_lib_ring_count;		/* kicad lib ring count */
 arnum = ustc_lib_arc_count;		/* kicad lib arc count */
 txnum = ustc_lib_text_count;		/* kicad lib text count */
 drnum = ustc_lib_draw_count;		/* kicad draw count */
 pnnum = ustc_lib_pin_count;		/* cad pin count */
 cnum = ustc_lib_comp_count;		/* kicad lib component count */
 buf[0] = '\0';
 if(ustc_lib_comp_read(			/* kicad lib components read */
    &cnum,				/* component number */
    &alnum,				/* alias number */
    &flnum,				/* field number */
    &drnum,				/* draw element number */
    &plnum,				/* poly number */
    &rcnum,				/* rectangle number */
    &rngnum,				/* ring number */
    &arnum,				/* arc number */
    &txnum,				/* text number */
    &pnnum, 				/* pin number */
    buf,				/* buffer */
    len					/* buffer length */
 ) == 0) {
    if(rvl == 1) {
	ustc_error("Kicad lib component read error\r\n");
	rvl = 0;
    }
 }
 ustc_lib_alias_count = alnum;		/* kicad lib alias count */
 ustc_lib_field_count = flnum;		/* kicad lib field count */
 ustc_lib_poly_count = plnum;		/* kicad lib point count */
 ustc_lib_rect_count = rcnum;		/* kicad lib rect count */
 ustc_lib_ring_count = rngnum;		/* kicad lib ring count */
 ustc_lib_arc_count = arnum;		/* kicad lib arc count */
 ustc_lib_text_count = txnum;		/* kicad lib text count */
 ustc_lib_draw_count = drnum;		/* kicad draw count */
 ustc_lib_pin_count = pnnum;		/* cad pin count */
 ustc_lib_comp_count = cnum;		/* kicad lib component count */
 return(rvl);
}
int ustc_lib_print(void)
{
 int rvl;
 rvl = 1;
 if(ustc_lib_header_print(&ustc_lib_header) == 0) {
    if(rvl == 1) {
	ustc_error("Kicac lib header print error\r\n");
	rvl = 0;
    }
 }
 if(ustc_lib_comps_print(ustc_lib_comp_count) == 0) {
    rvl = 0;
 }
 if(ustc_lib_comp_free() == 0) {	/* free lib comp */
    rvl = 0;
 }
 return(rvl);
}
int ustc_lib_recover(void)
{
 int rvl; 				/* return code */
 rvl = 1;				/* return code start */
 if(ustc_lib_read() == 0) {		/* lib read */
    rvl = 0;
 }
 if(ustc_lib_print() == 0) {		/* lib print */
    rvl = 0;
 }
 if(ustc_lib_free() == 0) {		/* lib free */
    rvl = 0;
 }
 return(rvl);
}
#define USTC_BS   	     0x7f	/* Back Space Key **/
#define USTC_DEL 	 0x335b1b	/* Delete Key **/
int ustc_key(u_int32_t key, u_char * buf, int len) /* read key, return key length bytes */
{
 if(buf == 0 || len == 0) {
    return(0);
 }
 if(key <= 0xFF) {			/* key length - one byte **/
    u_int8_t bchr;			/* buffer character **/
    bchr = *(buf + 0);			/* read characeter from buffer **/
    if(bchr == key) {			
        return(1);
    } else {
        return(0);			/* one byte key string not found in buffer **/
    }
 } else 
 if(key <= 0xFFFF) {			/* two bytes character **/
    if(len < 2) {			/* **/
	return(0);
    } else {
        u_int16_t bchr;			/* two bytes character **/
	bchr = 				/* calculate two bytes character **/
	    ( (((u_int32_t) *(buf + 1)) << 8) | 
	       ((u_int32_t) *(buf + 0)) );
	if(key == bchr) {		/* two bytes character found in buffer **/
	    return(2);
	} else {
	    return(0);
	} 
    }
 } else 
 if(key <= 0xFFFFFF) {			/* free bytes character **/
    if(len < 3) {
	return(0);
    } else {
        u_int32_t bchr;
	bchr = ( (((u_int32_t) *(buf + 2)) << 16) | 
	    (((u_int32_t) *(buf + 1)) << 8) | ((u_int32_t) *(buf + 0)) );
	if(key == bchr) {
	    return(3);
	} else {
	    return(0);
	}
    }
 } else {				/* for bytes character **/
    if(len < 4) {
	return(0);
    } else {
	u_int32_t bchr;
	bchr = 
	( (((u_int32_t) *(buf + 3)) << 24) | 
	  (((u_int32_t) *(buf + 2)) << 16) | 
	  (((u_int32_t) *(buf + 1)) << 8) | 
	   ((u_int32_t) *(buf + 0)) );
	if(key == bchr) {
	    return(4);
	} else {
	    return(0);
	}
    }
 }
}
/* echo = 0 - no echo, 1 - echo, 2 - '*' echo, 3 - '$' echo**/ 
/* wait - 0 = no wait, 1 = wait **/
int ustc_select(char * buf, int len, int echo, int wait)
{
 fd_set r_fds; 				/* read mask **/
 FD_ZERO(&r_fds);			/* clear read mask **/
/* read stdin file descriptor **/
 int stdinfd;				/* standart input file descriptor **/
 stdinfd = fileno(stdin);		/* read standart input file descriptor **/
 if((stdinfd < 0) || (stdinfd >= (int)sizeof(r_fds) * 8)) {
    ustc_error(				/* print error message **/
	"Stdin descriptor error {%d}\r\n",  /* format */
	stdinfd				/* stdin descriptor */
    );	
    return(-1);				/* error **/
 }
 int maxfd;				/* maximum descriptor number **/
 maxfd = stdinfd;			/* set maximal descriptor number **/
#define TC_GET_SET_ATTR
#ifdef TC_GET_SET_ATTR
 struct termios stdintios;		/* standart input attributes **/
 if(tcgetattr(stdinfd, &stdintios) < 0) { /* read attributes **/
    ustc_error(				/* print error message */
	"Set terminal attribute error {%d}\r\n",  /* format */
	errno				/* error code */
    );
    return(-1);
 }
 struct termios tios;			/* modified attributes **/
 tios = stdintios;			/* copy attributes **/
 cfmakeraw(&tios);			/* raw stdin **/
 if(
    tcsetattr(
	stdinfd, 			/* standart input file descriptor **/
	TCSANOW, 			/* set immedeatly **/
	&tios				/* raw atrtributes **/
    ) < 0) { 
    ustc_error(
	"Set terminal attribute error {%d}\r\n",
	errno
    );
    return(-1);
 }        
#endif 
 int stdinval;				/* stdin settings **/
 stdinval = fcntl(stdinfd ,F_GETFL,0);	/* read stdin settings **/
 if(fcntl(stdinfd, F_SETFL, O_NONBLOCK) < 0) {	
    ustc_error(				/* set non block mode error **/
	"Set nonblock mode for standart input error\n"
    );
    return(-1);				/* error **/
 }
 int pos;				/* destination buffer position **/
 int rvl;				/* return code **/
 pos = 0;				/* start buffer position **/
 while(pos < len) {			
    u_int8_t bch;			/* buffer character **/
    bch = *(buf + pos);			/* read buffer character **/
    if(bch == '\0') {			/* string separator **/
	break;
    }
    if(echo == 1) {
	ustc_printf("%c", bch);	/* print character **/
    } else 
    if(echo == 2) {
        ustc_printf("%c", '*');
    } else
    if(echo == 3) {
        ustc_printf("%c", '$');
    }
    pos += 1;				/* next buffer position **/
 }
 rvl = 1; 				/* start return value **/
 while(1) {
    FD_SET(
	stdinfd	,			/* stdin descriptor **/
	& r_fds				/* read descriptor **/
    );     
    struct timeval tval;		/* time interval **/
    if(wait > 0) {
        tval.tv_sec = wait;		/* 1sec delay **/
    } else {
        tval.tv_sec = 0;		/* no wait **/
    }
    tval.tv_usec = 0;			/* usec **/
    int cnt;				/* io events **/
    cnt = select(
        maxfd + 1,			/* descriptors in mask **/
	& r_fds, 			/* read mask **/
	0, 				/* write mask **/
	0, 				/* exeption mask **/
	& tval				/* time interval **/
    );
    if(kill_link > 1) {			/* received more then one term signal **/
	{ break; }			/* -> exit from loop **/
    }
    if(cnt < 0) {			/* select error **/
        if(
    	    (errno == EWOULDBLOCK) ||  /* opeation blocked **/
	    (errno == EINTR)
	) {	
	    continue;
	} else {
    	    rvl = -1;  		/* clear return code **/
	    ustc_error("Select function call error {%d}",  /* print error message **/
	        errno
	    );
    	    { break; }
	}
    } else
    if(cnt == 0) {			/* no events **/
        if(wait == 0) {
    	    { break; }			/* no wait -> exit from loop **/
	} else {
    	    continue;
	}
    }
    if(FD_ISSET(stdinfd,& r_fds) == 0) {
	continue;			/* not set standart input descriptor **/
    }
    u_char rbuf[32];			/* read buffer **/
    rbuf[0] = 0;
    while(1) {
	cnt = read(stdinfd, rbuf, sizeof(rbuf));
	if(cnt >= 0) {
	    { break; }
	}  
	if(
	    (errno == EWOULDBLOCK) || 	/* opeation blocked **/
	    (errno == EINTR)		/* operation iterrupted **/
	) {	
	    continue;
	} else {
    	    rvl = -1;  			/* read error**/
    	    ustc_error("Stdin read error {%d}\r\n", /* print error message **/
	        errno			/* errno number */
	    );
    	    { break; }
	}
    }
    if(rvl == -1) {
	{ break; }
    }
    if(kill_link > 1) {			/* receive more then one **/
	{ break; }			/* -> exit from loop **/
    } 
    if(cnt <= 0) {			/* empty buffer **/
        continue;			/* -> next read **/
    }
    int rpos = 0;			/* read postion **/
    while(rpos < cnt) {
        u_char pkey;			/* del or backspace key **/
        u_char plen;			/* key length **/
	plen = ustc_key(USTC_BS, rbuf, cnt);	/* check BS **/
	if(plen > 0) { 
	    pkey = 1;			/* BS key **/
	} else {
	    plen = ustc_key(USTC_DEL, rbuf, cnt); /* check DEL **/	
	    if(plen > 0) { 
	        pkey = 1;		/* DEL key **/
	    } else {
	        pkey = 0;		/* not DEL and not BS keys **/
	    }
	}
	if(pkey == 1) {
	    rpos += plen;		/* next buffer position **/
	    if(pos > 0) {
	        pos -= 1;		/* previous character position **/
	        if(buf != 0) {		/* not null buffer **/
	            if(pos < len) {	
	    	        *(buf + pos) = '\0'; /* write terminating character **/
		    }
		}
		if(
		    (echo == 1) || 	/* print character as is **/
		    (echo == 2) || 	/* print * as typed character **/
		    (echo == 3)		/* print # as typed character **/
		) {
		    ustc_printf("\b");
	    	    ustc_printf(" ");
		    ustc_printf("\b");
		}
	    }
	} else {
	    u_char rch;			/* readed character **/
	    rch = rbuf[rpos];		/* read character from buffer **/
	    if((rch == '\n') || (rch == '\r')) {
	        ustc_printf("\r\n");	/* enter **/
	        { break; }		/* string entered **/
	    } else
	    if((rch == 0x1B)) {		/* escape key **/
	        pos = 0;
	        ustc_printf("{Canceled}");
	        { break; }		/* string cancelled **/
	    } else {		/* valid character -> write character to buffer **/	
	        if(buf != 0) {		/* not null pointer **/
	    	    if(pos < len) {	/* position in buffer **/
		        if(pos >= 0) {	
			    *(buf + pos) = rch;
			}
		    }
		}
	        ++pos;			/* next buffer position **/
	        if(echo == 1) {
	    	    ustc_printf("%c",rch);
		} else 
		if(echo == 2) {
		    ustc_printf("%c",'*');
		} else
		if(echo == 3) {
		    ustc_printf("%c",'$');
		}
	    }
	    rpos += 1;    		/* next readed character **/
	} 
    }
    if(wait == 0) {			/* no wait mode **/
        { break; }			/* clear more **/
    }
    if(rpos < cnt) {			
        { break; }			/* read character processing terminating **/
    }
 }
 if(buf != 0) {				/* write string teminator **/
    if(pos > len) {
	pos = len - 1;
    }
    if(pos > 0) {
        *(buf + pos) = '\0';
    }
 }
 if(fcntl(stdinfd , F_SETFL, stdinval) < 0) {	
    ustc_error(				/* restore standart input **/
	"Restore setting for standart input error\n");
    return(-1);				/* error **/
 }
#ifdef TC_GET_SET_ATTR
 if(tcsetattr(stdinfd, TCSANOW, &stdintios) < 0) {
    ustc_error(
	"Set terminal attribute error {%d}\r\n", 
	errno
    );
    rvl = -1;
 }        
#endif 
 if(rvl < 0) {
    return(rvl);
 } else {
    return(pos);
 }
}
int ustc_lib_part(void)
{
 int rvl; 
 rvl = 1;
 if(ustc_lib_header_clear(&ustc_lib_header) == 0) {
    if(rvl == 1) {
	ustc_error("Kicac lib header clear error\r\n");
	rvl = 0;
    }
 }
 u_int8_t buf[256];
 u_int32_t len;
 len = sizeof(buf);
 if(ustc_lib_header_read(&ustc_lib_header, buf, len) == 0) {
    if(rvl == 1) {
	ustc_error("Kicac lib header read error\r\n");
	rvl = 0;
    }
 }
 u_int32_t cnum;
 u_int32_t alnum;
 u_int32_t flnum;
 u_int32_t drnum;
 u_int32_t plnum;
 u_int32_t rcnum;
 u_int32_t rngnum;
 u_int32_t arnum;
 u_int32_t txnum;
 u_int32_t pnnum;
 alnum = ustc_lib_alias_count;		/* kicad lib alias count */
 flnum = ustc_lib_field_count;		/* kicad lib field count */
 plnum = ustc_lib_poly_count;		/* kicad lib point count */
 rcnum = ustc_lib_rect_count;		/* kicad lib rect count */
 rngnum = ustc_lib_ring_count;		/* kicad lib ring count */
 arnum = ustc_lib_arc_count;		/* kicad lib arc count */
 txnum = ustc_lib_text_count;		/* kicad lib text count */
 drnum = ustc_lib_draw_count;		/* kicad draw count */
 pnnum = ustc_lib_pin_count;		/* cad pin count */
 cnum = ustc_lib_comp_count;		/* kicad lib component count */
 buf[0] = '\0';
 if(ustc_lib_comp_read(			/* kicad lib components read */
    &cnum,				/* component number */
    &alnum,				/* alias number */
    &flnum,				/* field number */
    &drnum,				/* draw element number */
    &plnum,				/* poly number */
    &rcnum,				/* rectangle number */
    &rngnum,				/* ring number */
    &arnum,				/* arc number */
    &txnum,				/* text number */
    &pnnum, 				/* pin number */
    buf,				/* buffer */
    len					/* buffer length */
 ) == 0) {
    if(rvl == 1) {
	ustc_error("Kicad lib component read error\r\n");
	rvl = 0;
    }
 }
 USTC_LIB_COMP * ulcri;			/* kicad source component pointer */
 u_int32_t crnum;			/* source component number */
 crnum = ustc_lib_comp_num(ustc_part);	/* component search */
 if(crnum != (u_int32_t)(-1)) {		/* reference componet number check */
    if(crnum >= ustc_lib_comp_count) {	/* out of component buffer check */
	ulcri = 0;			/* component pointer clear */
    } else {
	ulcri = ustc_lib_comp + crnum;	/* component pointer set */
    }
 } else {
    ulcri = 0;				/* component pointer clear */
 }
 if(ulcri != 0) {			/* reference component pointer check */
    if(ulcri->name != 0) {		/* reference component name pointer check */
	snprintf(			/* form part name */
	    (char *)ustc_part,		/* part name buffer */
	    sizeof(ustc_part),		/* part name buffer size */
	    "%s",			/* format */
	    ulcri->name			/* part name */
	);
    }
 }
 USTC_LIB_COMP ulci; 			/* lib component */
 if(ustc_lib_comp_clear(&ulci) == 0) {	/* lib component clear */
    ustc_error("Kicad lib component clear error\r\n");
    rvl = 0;
 }
 u_int8_t name[64];			/* name buffer */
 u_int8_t *pnam;			/* name pointer */
 u_int32_t size;			/* name buffer size */
 size = sizeof(name);			/* name buffer size calculate */
 ustc_printf("Name [%s] ", ustc_part);	/* name print */
 name[0] = '\0';			/* name clear */
 ustc_select((char *)name, size, 1, 1); /* name input */
 if(name[0] == '\0') {
    pnam = ustc_part;			/* name default */
 } else {
    pnam = name;			/* name entered */
 }
 len = strlen((char *) pnam) + 1;	/* component name length calculate */
 ulci.name = (u_int8_t *) malloc(len);	/* component name buffer allocate */
 if(ulci.name == 0) {			/* component name allocation check */
    if(rvl == 1) {
	ustc_error(			/* error message print */
	    "Component name {%s} allocation error\r\n", /* format */
	    pnam			/* component name pointer */
	);
	rvl = 0;			/* return code clear */
    }
 } else {
    snprintf(				/* component name form */
	(char *) ulci.name,		/* component name buffer */
	len,				/* component name buffer length */
	"%s",				/* format */
	pnam				/* component name pointer */
    );
 }
 ustc_printf("Reference ");		
 if(ulcri != 0) {			/* reference component pointer check */
    if(ulcri->reference != 0) {		/* ref component reference pointer check */ 
	ustc_printf("[%s] ", ulcri->reference); /* reference component reference print */
    } else {
	ustc_printf("[U] ");		/* reference default print */
    } 
 } else {
    ustc_printf("[U] ");		/* reference default print */
 }
 name[0] = '\0';			/* reference buffer clear */
 ustc_select((char *)name, size, 1, 1); /* reference input */
 if(name[0] == '\0') {			/* empty buffer check */
    if(ulcri != 0) {
	if(ulcri->reference != 0) {
	    pnam = ulcri->reference;	/* not null -> reference read */
	} else {
	    pnam = (u_int8_t *) "U";	/* reference default read */
	}
    } else {
	pnam = (u_int8_t *) "U";	/* reference default read */
    }
 } else {
    pnam = name;			/* reference name pointer set */
 }
 len = strlen((char *) pnam) + 1;	/* reference name length calculate */
 ulci.reference = (u_int8_t *) malloc(len); /* reference buffer allocate */
 if(ulci.reference == 0) {		/* reference buffer pointer check */
    if(rvl == 1) {			/* return code check */
	ustc_error(			/* error message print */
	    "Component reference {%s} allocation error\r\n", /* format */
	    pnam			/* reference name */
	);
	rvl = 0;			/* return code clear */
    }
 } else {
    snprintf(				/* form reference name */
	(char *) ulci.reference,	/* reference name buffer */
        len,				/* reference buffer length */
	"%s",				/* format */
        pnam				/* reference name buffer pointer */
    );
 }
 ustc_printf("Text offset ");		/* text offset print */
 if(ulcri != 0) {			/* reference component pointer check */
    ustc_printf("[%d] ", ulcri->offset); /* reference component offset print */
 } else {
    ustc_printf("[100] ");		/* default offset print */
 }
 name[0] = '\0';			/* input buffer clear */
 ustc_select((char *)name, size, 1, 1);	/* offset input */
 if(name[0] == '\0') {			/* buffer empty check */
    if(ulcri != 0) {			/* reference component pointer check */
	ulci.offset = ulcri->offset;	/* not null -> reference offset copy */
    } else {
	ulci.offset = 100;		/* component offset default write */
    }
 } else {
    ulci.offset = atoi((char *) name);	/* component ofset write */
 }
 ustc_printf("Display pin number ");	/* display pin number print */
 if(ulcri == 0) {			/* reference component pointer check */
    ustc_printf("[Y]/n ");		/* default print */
 } else {
    if(ulcri->pinnumvis > 0) {
	ustc_printf("[Y]/n ");
    } else {
	ustc_printf("y/[N] ");
    }
 }
 name[0] = '\0';			/* input buffer clear */
 ustc_select((char *)name, size, 1, 1);	/* offset input */
 if(name[0] == '\0') {			/* input offset check */
    if(ulcri != 0) {			/* reference component pointer check */
	ulci.pinnumvis = ulcri->pinnumvis; /* not null -> pin number visible copy */
    } else {
	ulci.pinnumvis = 1;		/* pin number visible set */
    }
 } else {
    if((name[0] == 'Y') || (name[0] == 'y')) {
	ulci.pinnumvis = 1;		/* pin number visible set */
    } else {
	ulci.pinnumvis = 0;		/* pin number visible clear */
    }
 }
 ustc_printf("Display pin name ");	/* display pin name print */
 if(ulcri == 0) {			/* reference component pointer check */
    ustc_printf("[Y]/n ");		/* default print */
 } else {
    if(ulcri->pinnamvis > 0) {		/* reference component pin name visible check */
	ustc_printf("[Y]/n ");
    } else {
	ustc_printf("y/[N] ");
    }
 } 
 name[0] = '\0';			/* input buffer clear */
 ustc_select((char *)name, size, 1, 1);	/* input */
 if(name[0] == '\0') {			/* input empty check */
    if(ulcri != 0) {			/* reference component pointer null check */
	ulci.pinnamvis = ulcri->pinnamvis; /* not null -> reference component pointer copy */
    } else {
	ulci.pinnamvis = 1;		/* pin name visible set */
    }
 } else {
    if((name[0] == 'Y') || (name[0] == 'y')) {
	ulci.pinnamvis = 1;		/* pin name visible set */
    } else {
	ulci.pinnamvis = 0;		/* pin name visible clear */
    }
 }
 ustc_printf("Unit count ");		/* unit count print */
 if(ulcri != 0) { 			/* reference component pointer check */
    ustc_printf("[%d] ", ulcri->units); /* not null -> reference component units print */
 } else {
    ustc_printf("[1] ");		/* default value print */
 }
 name[0] = '\0';			/* buffer clear */
 ustc_select((char *)name, size, 1, 1);	/* input */
 if(name[0] == '\0') {			/* buffer empty check */
    if(ulcri != 0) { 			/* reference component pointer check */
	ulci.units = ulcri->units;	/* reference units copy */
    } else {
	ulci.units = 1;			/* units default */
    } 
 } else {
    ulci.units = atoi((char *) name);	/* unit write */
 }
 ustc_printf("Unit can't be swapped ");	/* cannot be swapped */
 if(ulcri != 0) {			/* reference component pointer check */
    if(
	(ulcri->unitlock == USTC_LIB_COMP_UNITLOCK_FIXOLD) || /* fixol check */
	(ulcri->unitlock == USTC_LIB_COMP_UNITLOCK_FIX)	/* fix check */
    ) {
	ustc_printf("[N]/y ");
    } else {
	ustc_printf("[Y]/n ");
    }
 } else {
    ustc_printf("[N]/y ");
 }
 name[0] = '\0';			/* byffer clear */
 ustc_select((char *)name, size, 1, 1);	/* input */
 if(name[0] == '\0') {			/* buffer empty check */
    if(ulcri != 0) {			/* reference component pointer check */
	ulci.unitlock = ulcri->unitlock; /* not null -> reference lock copy */
    } else {
	ulci.unitlock = USTC_LIB_COMP_UNITLOCK_FIXOLD; /* spap enable write */
    }
 } else {
    if((name[0] == 'Y') || (name[0] == 'y')) {
	ulci.unitlock = USTC_LIB_COMP_UNITLOCK_LOCK; /* unit locked */
    } else {
	ulci.unitlock = USTC_LIB_COMP_UNITLOCK_FIXOLD; /* unit don't locked */
    }
 }
 ustc_printf("Power component ");	/* power component */
 if(ulcri != 0) {			/* reference component pointer check */
    if(ulcri->power > 0) {		/* reference component power check */
	ustc_printf("[Y]/n ");
    } else {
	ustc_printf("y/[N] ");
    }
 } else {
    ustc_printf("[N]/y ");
 }
 name[0] = '\0';			/* buffer clear */
 ustc_select((char *)name, size, 1, 1);	/* input */
 if(name[0] == '\0') {			/* buffer empty check */
    if(ulcri != 0) {			/* reference component pointer check */
	ulci.power = ulcri->power;	/* reference component power copy */
    } else {
	ulci.power = 0;			/* default write */
    }
 } else {
    if((name[0] == 'Y') || (name[0] == 'y')) {
	ulci.power = 1;
    } else {
	ulci.power = 0;
    }
 }
 ulci.alias = alnum;			/* alias reference */
 ulci.alcount = 0;			/* alias count */
 u_int32_t alidx;			/* alias index */
 alidx = 0;				/* alias index start value */
 while(1) { 				/* alias loop */
    ustc_printf("Alias,");
    pnam = name;			/* alias pointer defalut */
    if(ulcri != 0) {			/* reference component pointer check */
	if(alidx < ulcri->alcount) {	/* alias index check */
	    u_int32_t alref;		/* reference alias number */
	    alref = ulcri->alias + alidx; /* reference alias number calculate */	
	    if(alref < ustc_lib_alias_count) { /* alias number check */
		USTC_LIB_ALIAS * ulai;	/* alias pointer */
		ulai = ustc_lib_alias + alref; /* alias pointer set */
		pnam = ulai->name;	/* alias read */
		if(alidx > 0) {
		    ustc_printf(",");
		}
		ustc_printf("[%s]", pnam);
	    }
	    ustc_printf(".");
	}
	alidx += 1;			/* next alias index */
    }
    ustc_printf("Enter to terminate ");
    name[0] = '\0';			/* buffer clear */
    ustc_select((char *)name, size, 1, 1); /* input */
    if(name[0] == '\0') {		/* input default check */
	if(*(pnam + 0) == 0) {		/* input empty check */
	    { break; }
	}
    } else {
	pnam = name;
    }
    if(ustc_lib_alias_add(		/* kicad lib alias add */
	&alnum,				/* alias number */
	pnam				/* alias name */
    ) == 0) {
	if(rvl == 1) {			/* return code check */
	    ustc_error(			/* error message print */
		"Component alias {%s} add error\r\n",
		pnam			/* alias */
	    );
	    rvl = 0;			/* return code clear */
	    { break; }			/* loop exit */
	}
    }
 }
 ulci.alcount = alnum - ulci.alias;	/* alias count */
 u_int32_t pincount;			/* component pincount */
 ustc_printf("Pin count ");		/* pin count message print */
 if(ulcri != 0) {			/* lib component pointer check */
    ustc_printf("[%d] ", ulcri->pcount); /* reference component pin count print */
 } else {
    ustc_printf("[14] ");		/* default pin count print */	
 } 
 name[0] = '\0';			/* buffer clear */
 ustc_select((char *)name, size, 1, 1);	/* buffer input */
 if(name[0] == '\0') {			/* empty buffer check */
    if(ulcri != 0) {			/* referencfe component pointer check */
	pincount = ulcri->pcount;	/* reference pin count read */
    } else {
	pincount = 14;			/* default pin count */
    }
 } else {
    pincount = atoi((char *) name);	/* pin count read */
 }
 if(pincount < 0) {
    pincount = 0;
 }
 u_int32_t width;
 u_int32_t height;
 width = strlen((char *) ulci.name) * 60;
 len = strlen((char *) ulci.reference) * 60;
 if(width < len) {
    width = len;
 }
 if(width < (u_int32_t)(3 * (ulci.offset * 3))) {
    width = (3 * (ulci.offset * 3));
 }
 u_int32_t lwidth;
 u_int32_t rwidth;
 lwidth = 0;
 rwidth = 0;
 height = (pincount / 2) * 100 + 2 * ulci.offset;
 USTC_LIB_RECT ulri;
 ulri.x0 = -(width / 2);		/* x0 point */
 ulri.y0 = -(height / 2);		/* y0 point */
 ulri.x1 =  (width / 2);		/* x0 point */
 ulri.y1 =  (height / 2);			/* y0 point */
 ulri.fill = 0;				/* rect filling */
 USTC_LIB_DRAW uldi;			/* kicad lib draw definition */
 uldi.type = USTC_LIB_DRAW_RECT;	/* draw segment type */
 uldi.unit = 0;				/* 0 - common in all parts, else part number */
 uldi.convert = 0;			/* 0 - common in 2 representations, else 1 or 2 */
 uldi.width = 1; 			/* line width */
 uldi.ref = rcnum; 			/* draw reference */
 ulci.draw = drnum;			/* pin reference */
 ulci.pin = pnnum;			/* pin reference */
 ulci.pcount = 0;			/* pin count */
 u_int32_t pidx;			/* pin index */
 u_int32_t pcnt;			/* pin count */
 u_int32_t piny;			/* pin y coordinate */
 pidx = 1;				/* pin index */
 pcnt = 0;				/* pin count */ 
 piny = (height / 2) - 2 * ulci.offset;	/* pin y */
 while(pcnt < pincount) {		/* pin loop */
    pcnt += 1; 				/* pin entered */
    USTC_LIB_PIN ulpi;			/* lib pin */
    if(ustc_lib_pin_clear(&ulpi) == 0) { /* pin clear */
	ustc_error("Kicad lib pin clear error\r\n");
	return(0);
    }
    ustc_printf("Pin number %d ", pidx); /* pin index print */
    name[0] = '\0';			/* name clear */
    ustc_select((char *)name, size, 1, 1); /* input */
    if(name[0] != '\0') {		/* input empty check */
	pidx = atoi((char *) name);	/* index read */
    }
    if(pidx < 1) {			/* pin index check */
	pidx = 1;			/* pin index set low limit */
    }
    USTC_LIB_PIN * ulpri;		/* reference pin pointer */
    ulpri = 0;				/* reference pin pointer clear */
    if(ulcri != 0) {			/* reference component pointer check */
	u_int32_t rpidx;		/* reference pin index */	
	for(rpidx = 0; rpidx < ulcri->pcount; rpidx += 1) { /* reference pin loop */
	    USTC_LIB_PIN * ulpwi;	/* pin work pointer */
	    u_int32_t pref;		/* pin reference */
	    pref = ulcri->pin + rpidx;	/* pin reference number calculate */
	    if(pref < ustc_lib_pin_count) { /* pin reference check */
		ulpwi = ustc_lib_pin + pref; /* wroek pointer set */
		if(ulpwi->num == pidx) { /* work pin number check */
		    ulpri = ulpwi;	/* reference pin pointer set */
		    { break; }		/* loop exit */
		}
	    } else {
		if(rvl == 1) {		/* return code check */
		    ustc_error(		/* error message print */
			"Kicad lib pin out of buffer,component name {%s},"
			"component start pin {%d}, component pin count {%d},"
			"pin count {%d}\r\n",
			ulcri->name,	/* component name */
			ulcri->pin,	/* component pin start number */
			ulcri->pcount,	/* component pin count */
			ustc_lib_pin_count /* library pin count */
		    );
		    rvl = 0;		/* return code clear */
		}
	    }
	}
    }
    ustc_printf("Pin name ");		/* pin name print */
    if(ulpri != 0) {			/* reference pin pointer check */
	ustc_printf("[%s] ", ulpri->name); /* reference pin name print */
    } else {
	ustc_printf("[%d] ", pidx);	/* default pin name print */
    }
    name[0] = '\0';			/* buffer clear */
    pnam = name;			/* name pointer */
    ustc_select((char *)name, size, 1, 1); /* input */
    if(name[0] == '\0') {		/* default input check */
	if(ulpri != 0) {		/* reference pointer check */
	    pnam = ulpri->name;		/* pin name pinter set */
	    if(pnam == 0) {		/* pin name pointer check */
		pnam = name;		/* pin name pointer reset to buffer */
		snprintf(		/* form pin name */
		    (char *) name,	/* name buffer */
		    sizeof(name),	/* name buffer size */
		    "%d",		/* format */
		    pidx		/* pin index */
		);
	    }
	} else {
	    snprintf(			/* form pin name */
		(char *) name,		/* pin name buffer */
		sizeof(name),		/* pin name buffer size */
	        "%d",			/* format */
		pidx			/* index */
	    );
	}
    }
    len = strlen((char *) pnam) + 1;	/* pin name length calculate */
    ulpi.name = (u_int8_t *) malloc(len); /* pin name buffer allocate */
    if(ulpi.name == 0) {		/* name allocation check */
	ustc_error(			/* error print */
	    "Pin name {%s} allocation error\r\n", /* format */
	    pnam			/* pin name */
	);
	rvl = 0;			/* return code clear */
	{ break; }			/* loop exit */
    } else {
	snprintf(			/* form pin name */
	    (char *) ulpi.name,		/* pin name buffer */
	    len,			/* buffer length */
	    "%s",			/* format */
	    pnam			/* pin name pointer */
	);
	pnam = ulpi.name;
	while(*pnam != '\0') {		/* string terminator check */
	    if(*pnam == ' ') {		/* space check */
		*pnam = '_';
	    }
	    pnam += 1;			/* next character */
	}
    }
    ulpi.num = pidx;			/* pin number */
    pidx += 1;				/* pin index next */
    u_int32_t wid;
    wid = strlen((char *) ulpi.name) * 60;
    if(pcnt <= pincount / 2) {
	ulpi.x0 = -(width / 2) - (3 * ulci.offset); /* x position */
	if(wid > lwidth) {
	    lwidth = wid;
	} 
    } else {
	ulpi.x0 = +(width / 2) + (3 * ulci.offset); /* x position */
	if(wid > lwidth) {
	    rwidth = wid;
	}
    }
    if(pcnt == (pincount / 2) + 1) {
	piny = (height / 2) - ulci.offset;
    }
    ulpi.y0 = piny;			/* pin y write */
    piny -= 100;			/* pin y next calculate */
    ulpi.length = (3 * ulci.offset);	/* pin length */
    if(pcnt <= pincount / 2) {
	ulpi.ornt = USTC_LIB_PIN_ORNT_RIGHT; /* pin orientation */
    } else {
	ulpi.ornt = USTC_LIB_PIN_ORNT_LEFT; /* pin orientation */
    }
    ulpi.numheight = 60;		/* hunber height */
    ulpi.nameheight = 60;		/* name height */
    ulpi.unit = 0;			/* part count */
    ulpi.convert = 0;			/* 0 - common in 2 representations, else 1 or 2 */
    u_int8_t * iestr;			/* input string */
    u_int8_t * oestr;			/* output string */
    u_int8_t * bestr;			/* bidirectional string */
    u_int8_t * testr;			/* tristate string */
    u_int8_t * pestr;			/* passive string */
    u_int8_t * uestr;			/* unspecified string */
    u_int8_t * piestr;			/* power in string */
    u_int8_t * poestr;			/* power out string */
    u_int8_t * cestr;			/* collector string */
    u_int8_t * eestr;			/* emitor string */
    iestr = (u_int8_t *)"I";		/* input string not selected */
    oestr = (u_int8_t *)"O";		/* output string not selected */
    bestr = (u_int8_t *)"B";		/* bidirectional string not selected */
    testr = (u_int8_t *)"T";		/* tristate string not selected */
    pestr = (u_int8_t *)"P";		/* passive string not selected */
    uestr = (u_int8_t *)"U";		/* unspecified string not selected */
    piestr = (u_int8_t *)"W";		/* power string not selected */
    poestr = (u_int8_t *)"w";		/* power string not selected */
    cestr = (u_int8_t *)"C";		/* colector string not selected */
    eestr = (u_int8_t *)"E";		/* emitor string not selected */
    if(ulpri != 0) {			/* reference pin pointer check */
	if(ulpri->etype == USTC_LIB_PIN_ETYPE_INPUT) { /* type input check */
	    iestr = (u_int8_t *)"[I]";	/* input string selected */
	} else
	if(ulpri->etype == USTC_LIB_PIN_ETYPE_OUTPUT) { /* type output check */
	    oestr = (u_int8_t *)"[O]";	/* otput string selected */
	} else
	if(ulpri->etype == USTC_LIB_PIN_ETYPE_BIDI) { /* type bidirectinal check */
	    bestr = (u_int8_t *)"[B]";	/* bidirectional string selected */
	} else
	if(ulpri->etype == USTC_LIB_PIN_ETYPE_TRISTATE) { /* type tristate check */
	    testr = (u_int8_t *)"[T]";	/* tristate string selected */
	} else
	if(ulpri->etype == USTC_LIB_PIN_ETYPE_PASSIVE) { /* type passive check */
	    pestr = (u_int8_t *)"[P]";	/* passive string selected */
	} else
	if(ulpri->etype == USTC_LIB_PIN_ETYPE_UNSPECIFIED) { /* type unspecified check */
	    uestr = (u_int8_t *)"[U]";	/* unspecified string selected */
	} else
	if(ulpri->etype == USTC_LIB_PIN_ETYPE_POWER_IN) { /* type power in check */
	    piestr = (u_int8_t *)"[W]"; /* pwer in string selected */
	} else
	if(ulpri->etype == USTC_LIB_PIN_ETYPE_POWER_OUT) { /* type power out check */
	    poestr = (u_int8_t *)"[w]"; /* power out string selected */
	} else
	if(ulpri->etype == USTC_LIB_PIN_ETYPE_OPEN_COLLECTOR) { /* type open collector check */
	    cestr = (u_int8_t *)"[C]";	/* open collector string selected */
	} else
	if(ulpri->etype == USTC_LIB_PIN_ETYPE_OPEN_EMITOR) { /* type open emitor check */
	    eestr = (u_int8_t *)"[E]";	/* open emitor selected */
	} else {
	    iestr = (u_int8_t *)"[I]";	/* input string selected */
	}
    } else {
	iestr = (u_int8_t *)"[I]";	/* input string selected */
    }
    ustc_printf(
	"Pin etype %s-Input,%s-Output,%s-Bibidirectional,%s-Tristate,%s-Passive,\r\n"
	"%s-Unspec,%s-power in,%s-power out,%s-open collector,%s-open emitor ",
	iestr,				/* input string */
	oestr,				/* output string */
	bestr,				/* bidirectional string */
	testr,				/* tristate string */
	pestr,				/* passive string */
	uestr,				/* unspecified string */
	piestr,				/* power in string */
	poestr,				/* power out string */
	cestr,				/* collector string */
	eestr				/* emitor string */
    );
    name[0] = '\0';			/* buffer clear */
    ustc_select((char *)name, size, 1, 1); /* input */
    if(name[0] == '\0') {		/* empty input check */
	if(ulpri != 0) {		/* reference pin pointer check */
	    ulpi.etype = ulpri->etype;	/* reference electrical type copy */
	} else {
	    ulpi.etype = USTC_LIB_PIN_ETYPE_INPUT; /* default electrical type write */
	}
    } else {
	if((name[0] == 'I') || (name[0] == 'i')) { /* input check */
	    ulpi.etype = USTC_LIB_PIN_ETYPE_INPUT; /* input write */
	} else
        if((name[0] == 'O') || (name[0] == 'o')) { /* output check */
	    ulpi.etype = USTC_LIB_PIN_ETYPE_OUTPUT; /* output write */
	} else
	if((name[0] == 'B') || (name[0] == 'b')) { /* bidirectional check */
	    ulpi.etype = USTC_LIB_PIN_ETYPE_BIDI; /* bidirectional write */
	} else
	if((name[0] == 'T') || (name[0] == 't')) { /* tristate check */
	    ulpi.etype = USTC_LIB_PIN_ETYPE_TRISTATE; /* tristate write */
	} else
	if((name[0] == 'P') || (name[0] == 'p')) { /* passive check */
	    ulpi.etype = USTC_LIB_PIN_ETYPE_PASSIVE; /* passive write */
	} else
	if((name[0] == 'U') || (name[0] == 'u')) { /* unspecified check */
	    ulpi.etype = USTC_LIB_PIN_ETYPE_UNSPECIFIED; /* unspecified write */
	} else
	if((name[0] == 'W')) {		/* power in check */
	    ulpi.etype = USTC_LIB_PIN_ETYPE_POWER_IN; /* power in write */
	} else
	if((name[0] == 'w')) {		/* power out check */
	    ulpi.etype = USTC_LIB_PIN_ETYPE_POWER_OUT; /* power out write */
	} else
	if((name[0] == 'C') || (name[0] == 'c')) {  /* open collector check */
	    ulpi.etype = USTC_LIB_PIN_ETYPE_OPEN_COLLECTOR; /* open collector write */
	} else
	if((name[0] == 'E') || (name[0] == 'e')) { /* open emitor check */
	    ulpi.etype = USTC_LIB_PIN_ETYPE_OPEN_EMITOR; /* open emitor write */
	} else {
	    ulpi.etype = USTC_LIB_PIN_ETYPE_INPUT;
	}
    }
    u_int8_t * lsstr;			/* line shape string */
    u_int8_t * isstr;			/* inverce shape string */
    u_int8_t * csstr;			/* clock shape string */
    u_int8_t * icsstr;			/* invert clock string */
    u_int8_t * lisstr;			/* low input string */
    u_int8_t * lcsstr;			/* low clock string */
    u_int8_t * vsstr;			/* low output string */
    lsstr = (u_int8_t *)"L";		/* line shape string not selected */
    isstr = (u_int8_t *)"I";		/* inverce shape string not selected */
    csstr = (u_int8_t *)"C";		/* clock shape string not selected */
    icsstr = (u_int8_t *)"IC";		/* invert clock string not selected */
    lisstr = (u_int8_t *)"LI";		/* low input string not selected */
    lcsstr = (u_int8_t *)"LC";		/* low clock string not selected */
    vsstr = (u_int8_t *)"V";		/* low output string not selected */
    if(ulpri != 0) { 			/* reference pin pointer check */
	if(ulpri->shape == USTC_LIB_PIN_SHAPE_LINE) { /* line shape check */
	    lsstr = (u_int8_t *)"[L]";	/* line string selected */
	} else
	if(ulpri->shape == USTC_LIB_PIN_SHAPE_INVERT) { /* invert shape check */
	    isstr = (u_int8_t *)"[I]";	/* invert shape string selected */
	} else
	if(ulpri->shape == USTC_LIB_PIN_SHAPE_CLOCK) { /* clock shape check */
	    csstr = (u_int8_t *)"[C]";	/* clock shape string selected */
	} else
	if(ulpri->shape == USTC_LIB_PIN_SHAPE_CLOCK_INV) { /* clock invert shape check */
	    icsstr = (u_int8_t *)"[IC]"; /* invert clock string selected */
	} else
	if(ulpri->shape == USTC_LIB_PIN_SHAPE_LOW_IN) { /* low in shape check */
	    lisstr = (u_int8_t *)"[LI]";/* low in shape string selected */
	} else
 	if(ulpri->shape == USTC_LIB_PIN_SHAPE_LOW_CLOCK) { /* low clock shape check */
	    lcsstr = (u_int8_t *)"[LC]"; /* low clock string selected */
	} else
	if(ulpri->shape == USTC_LIB_PIN_SHAPE_LOW_OUT) { /* low out shape check */
	    vsstr = (u_int8_t *)"[V]";	/* low otput string selected */
	} else {
	    lsstr = (u_int8_t *)"[L]";	/* line string selected */
	}
    } else {
	lsstr = (u_int8_t *)"[L]";	/* line string selected */
    }
    ustc_printf(
	"Pin shape %s-Line,%s-invert,%s-clock,%s-clock invert,%s-low in,\r\n"
	"%s-low clock,%s-low out ",
	lsstr,				/* line string */
        isstr,				/* invert string */
	csstr,				/* clock */
        icsstr,				/* invert clock */
	lisstr,				/* low input */
        lcsstr,				/* low clock */
	vsstr				/* low output */
    );
    name[0] = '\0';			/* buffer clear */
    ustc_select((char *)name, size, 1, 1); /* name clear */
    if(name[0] == '\0') {		/* input empty check */
	if(ulpri != 0) { 		/* reference pin pointer check */
	    ulpi.shape = ulpri->shape;	/* shape copy */
	} else {
	    ulpi.shape = USTC_LIB_PIN_SHAPE_LINE; /* shape default write */
	}
    } else {
        if(
	    (name[0] == '\0') || 	/* empty input check */
	    (
		((name[0] == 'L') || (name[0] == 'l')) &&
		 (name[1] == '\0')
	    )
	) {
	    ulpi.shape = USTC_LIB_PIN_SHAPE_LINE; /* line */
	} else
        if(
    	    ((name[0] == 'I') || (name[0] == 'i')) &&
	     (name[1] == '\0')
	) {
	    ulpi.shape = USTC_LIB_PIN_SHAPE_INVERT; /* invert */
	} else
	if(
	    ((name[0] == 'C') || (name[0] == 'c')) && /* clock */
	     (name[1] == '\0')
	) { 
	    ulpi.shape = USTC_LIB_PIN_SHAPE_CLOCK; /* clock */
	} else
	if(
	    ((name[0] == 'I') || (name[0] == 'i')) &&
	    ((name[1] == 'C') || (name[1] == 'c')) &&
	     (name[2] == '\0')
	) {
	    ulpi.shape = USTC_LIB_PIN_SHAPE_CLOCK_INV; /* clock invert */
	} else
	if(
	    ((name[0] == 'L') || (name[0] == 'l')) &&
	    ((name[1] == 'I') || (name[1] == 'i')) &&
	     (name[2] == '\0')
	) {
	    ulpi.shape = USTC_LIB_PIN_SHAPE_LOW_IN; /* low in */
	} else
	if(
	    ((name[0] == 'L') || (name[0] == 'l')) &&
	    ((name[1] == 'C') || (name[1] == 'c')) &&
	     (name[2] == '\0')
	) {
	    ulpi.shape = USTC_LIB_PIN_SHAPE_LOW_CLOCK; /* low clock */
	} else
	if((name[0] == 'V') || (name[0] == 'v')) {
	    ulpi.shape = USTC_LIB_PIN_SHAPE_LOW_OUT; /* low out */
	} else {
	    ulpi.shape = USTC_LIB_PIN_SHAPE_LINE;
	}
    }
    if(ustc_lib_pin_add(			/* kicad lib rect add */
	&pnnum,
	&ulpi
    ) == 0) {
	ustc_error("Pin add error\r\n");
	return(0);
    }
 }
 ulci.pcount = pnnum - ulci.pin;	/* pin count */
 if(lwidth + rwidth + (ulci.offset * 3) > width) {
    width = lwidth + rwidth + (ulci.offset * 3); 
    width = (width + 99) / 100;
    width = (width * 100);
    ulri.x0 = -(width / 2);		/* x0 point */
    ulri.x1 =  (width / 2);		/* x0 point */
    for(pidx = 0; pidx < ulci.pcount; ++pidx) { /* pin index */
	u_int32_t pref;			/* pin reference */
	pref = ulci.pin + pidx;		/* pin reference calculate */
	if(pref >= ustc_lib_pin_count) { /* pin reference out of buffer check */
	    if(rvl == 1) {		/* return code check */
		ustc_error(		/* error message print */
		    "Out of pin buffer error, component name {%s},"
		    "start pin {%s}, component pin count {%d},"
		    "pin count {%d}\r\n",
		    ulci.name,		/* component name */
		    ulci.pin,		/* pin number */
		    ulci.pcount,	/* component pin count */
		    ustc_lib_pin_count	/* pin count */
		);
		rvl = 0;		/* return code clear */
	    }
	    continue;
	} 
	USTC_LIB_PIN * ulpi;		/* pin pointer */
	ulpi = ustc_lib_pin + pref;	/* pin pointer set */
	if(pidx < ulci.pcount / 2) {
	    ulpi->x0 = -(width / 2) - (3 * ulci.offset); /* x position */
	} else {
	    ulpi->x0 = +(width / 2) + (3 * ulci.offset); /* x position */
	}
    }
 }
 ulci.ref = flnum;			/* field reference */
 ulci.val = (flnum + 1);		/* field value */
 USTC_LIB_FIELD ulfi;			/* kicad lib field */
 if(ustc_lib_field_clear(&ulfi) == 0) {
    if(rvl == 1) {
	ustc_error("Lib field clear error\r\n");
	rvl = 0;
    }
 }
 len = strlen((char *)ulci.name) + 1;
 ulfi.name = (u_int8_t *) malloc(len);	/* text name */
 if(ulfi.name == 0) {
    ustc_error(
	"Componenht name {%s} allocation error\r\n",
	ulci.reference
    );
    rvl = 0;
 } else {
    snprintf(
	(char *)ulfi.name,
	len,
	"%s",
	ulci.reference
    );
 }
 ulfi.x0 = (width / 2);			/* x position, mills */
 ulfi.y0 = (height / 2);		/* y position, mills */
 ulfi.height = 60;			/* text height */
 ulfi.ornt = USTC_LIB_FIELD_ORNT_HOR;	/* field orientation */
 ulfi.vis = 1;				/* field visible */
 ulfi.hor = USTC_LIB_FIELD_HOR_R;	/* horizontal justify */
 ulfi.vert = USTC_LIB_FIELD_VERT_B;			/* vertical justify unknown */	
 if(ustc_lib_field_add(
	&flnum,
	&ulfi
 ) == 0) {
    ustc_error(
	"Field add error\r\n"
    );
    rvl = 0;
 }
 if(ustc_lib_field_clear(&ulfi) == 0) {
    if(rvl == 1) {
	ustc_error("Lib field clear error\r\n");
	rvl = 0;
    }
 }
 len = strlen((char *)ulci.name) + 1;
 ulfi.name = (u_int8_t *) malloc(len);	/* text name */
 if(ulfi.name == 0) {
    ustc_error(
	"Componenht name name {%s} allocation error\r\n",
	ulci.name
    );
    rvl = 0;
 } else {
    snprintf(
	(char *)ulfi.name,
	len,
	"%s",
	ulci.name
    );
 }
 ulfi.x0 = (width / 2);			/* x position, mills */
 ulfi.y0 = -(height / 2)-100;		/* y position, mills */
 ulfi.height = 60;			/* text height */
 ulfi.ornt = USTC_LIB_FIELD_ORNT_HOR;	/* field orientation */
 ulfi.vis = 1;				/* field visible */
 ulfi.hor = USTC_LIB_FIELD_HOR_R;	/* horizontal justify */
 ulfi.vert = USTC_LIB_FIELD_VERT_B;			/* vertical justify unknown */	
 if(ustc_lib_field_add(
	&flnum,
	&ulfi
 ) == 0) {
    ustc_error(
	"Field add error\r\n"
    );
    rvl = 0;
 }
 if(ustc_lib_rect_add(			/* kicad lib rect add */
	&rcnum,
	&ulri
 ) == 0) {
    ustc_error("Rectangle add error\r\n");
    return(0);
 }
 if(ustc_lib_draw_add(			/* kicad lib rect add */
	&drnum,
	&uldi
 ) == 0) {
    ustc_error("Draw add error\r\n");
    return(0);
 }
 ulci.dcount = (drnum - ulci.draw);	/* pin count */
 if(ustc_lib_comp_add(
    &cnum,
    &ulci
 ) == 0) {
    ustc_error(
        "Component {%s} add error\r\n",
        ulci.name
    );
    return(0);
 }
 ustc_lib_alias_count = alnum;		/* kicad lib alias count */
 ustc_lib_field_count = flnum;		/* kicad lib field count */
 ustc_lib_poly_count = plnum;		/* kicad lib point count */
 ustc_lib_rect_count = rcnum;		/* kicad lib rect count */
 ustc_lib_ring_count = rngnum;		/* kicad lib ring count */
 ustc_lib_arc_count = arnum;		/* kicad lib arc count */
 ustc_lib_text_count = txnum;		/* kicad lib text count */
 ustc_lib_draw_count = drnum;		/* kicad draw count */
 ustc_lib_pin_count = pnnum;		/* cad pin count */
 ustc_lib_comp_count = cnum;		/* kicad lib component count */
 if(ustc_lib_header_print(&ustc_lib_header) == 0) {
    if(rvl == 1) {
	ustc_error("Kicac lib header print error\r\n");
	rvl = 0;
    }
 }
 if(ustc_lib_comps_print(crnum) == 0) {
    rvl = 0;
 }
 if(ustc_lib_comp_free() == 0) {	/* free lib comp */
    rvl = 0;
 }
 return(rvl);
}
USTC_PADSNET_COMP * ustc_padsnet_comp = 0; /* padsnet component buffer */
u_int32_t ustc_padsnet_comp_count = 0;	/* padsnet component count */
USTC_PADSNET_NODE * ustc_padsnet_node = 0; /* padsnet node buffer */
u_int32_t ustc_padsnet_node_count = 0;	/* padsnet node count */
USTC_PADSNET_SIGNAL * ustc_padsnet_signal = 0;
u_int32_t ustc_padsnet_signal_count = 0;
int ustc_padsnet_recover(void)
{
 int rvl;
 rvl = 1;
 if(ustc_padsnet_read() == 0) {
    rvl = 0;
 }
 if(ustc_padsnet_print() == 0) {
    rvl = 0;
 }
 if(ustc_padsnet_free() == 0) {
    rvl = 0;
 }
 return(rvl);
}
int main(int argc, char * args[])
{					/* set signal functions **/
 int rvl;
 rvl = 1;
/* 1 - 16 **/
 int inum;				/* signal definition index **/	
 inum = 0;				/* start index **/
 while(1) {
    struct {
	int snum;				/* signal number **/
	void (*sfun)(int, siginfo_t *, void *); /* signal function **/
    } sigdef[] = {
	{ SIGHUP   , ustc_bad_signal_ext },	/* 1 - Hangup **/
	{ SIGINT   , ustc_bad_signal_ext },	/* 2 - Interrupt **/
	{ SIGQUIT  , ustc_bad_signal_ext },	/* 3 - Quit **/
	{ SIGILL   , ustc_bad_signal_ext },	/* 4 - Illegal instruction **/
	{ SIGTRAP  , ustc_bad_signal_ext },	/* 5 - trap **/
	{ SIGABRT  , ustc_bad_signal_ext },	/* 6 - abort **/
	{ SIGBUS   , ustc_bad_signal_ext },	/* 7 - buss error **/
	{ SIGFPE   , ustc_bad_signal_ext },	/* 8 - floating point exeption **/
/* not processed signal **/ 
/*	{ SIGKILL  , ustc_bad_signal_ext }, **/	/* 9 - kill **/
	{ SIGUSR1  , ustc_bad_signal_ext },	/* 10 - user signal 1 **/
	{ SIGSEGV  , ustc_bad_signal_ext },	/* 11 - segmetation violation **/
	{ SIGUSR2  , ustc_bad_signal_ext },	/* 12 - user signal 2 **/
	{ SIGPIPE  , ustc_sig_pipe_ext },	/* 13 - pipe **/
	{ SIGALRM  , ustc_sig_alarm_ext },	/* 14 - alarm **/
	{ SIGTERM  , ustc_sig_term_ext },	/* 15 - term **/
	{ SIGSTKFLT, ustc_bad_signal_ext },	/* 16 - stack failt **/
/* 17-31 **/
	{ SIGCHLD  , ustc_sig_chld_ext },	/* 17 - children **/
	{ SIGCONT  , ustc_bad_signal_ext },	/* 18 - continue **/
/* not processed signal **/ 
/*	{ SIGSTOP  , ustc_bad_signal_ext }, **/	/* 19 - stop **/
	{ SIGTSTP  , ustc_bad_signal_ext },	/* 20 - terminal stop, terminal command **/
	{ SIGTTIN  , ustc_bad_signal_ext },	/* 21 - terminal read **/
	{ SIGTTOU  , ustc_bad_signal_ext },	/* 22 - terminal weite **/
	{ SIGURG   , ustc_bad_signal_ext },	/* 23 - urg **/
	{ SIGXCPU  , ustc_bad_signal_ext },	/* 24 - cpu limit exceeded **/
	{ SIGXFSZ  , ustc_bad_signal_ext },	/* 25 - file size exceeded **/
	{ SIGVTALRM, ustc_bad_signal_ext },	/* 26 - virtual alarm clock **/
	{ SIGPROF  , ustc_bad_signal_ext },	/* 27 - profilling alarm clock **/
	{ SIGWINCH , ustc_bad_signal_ext },	/* 28 - window size changed **/
	{ SIGIO    , ustc_bad_signal_ext },	/* 29 - I/O **/
	{ SIGPWR   , ustc_bad_signal_ext },	/* 30 - power failure restart **/
	{ SIGSYS   , ustc_bad_signal_ext },	/* 31 - bad system call **/
/* 32 **/
/* 33 **/
/* 34 **/
	{ 0, 0 }
    };
    int snum;				/* signal number **/
    snum = sigdef[inum].snum;		/* read signal number **/
    if(snum == 0) {			/* signal array terminator **/
	{ break; }			/* terminate loop **/
    }
    /* set signal process function **/
    if(ustc_signal(snum,sigdef[inum].sfun) == 0) {
    /* set signal error -> clear return code **/
        rvl = 0;
    }
    inum += 1;				/* next index in signal description **/
 }
/* 35 - 64 **/
 {
    int snum;				/* signal number **/
    for(snum = SIGRTMIN; snum <	SIGRTMAX; ++snum) {	
    /* set run time signals 35 - 64 **/
	if(ustc_signal(snum,ustc_bad_signal_ext) == 0) {
	/* set signal error -> clear return code **/
	    rvl = 0;
	}
    }
 }
/* search in input args help request **/
 int argn;
 if(argc >= 1) {
    if( (*(args[0] + 0) == '.') || (*(args[0] + 1) == '/') ) {
	ustc_local = 1;			/* local path **/
    } else {
	ustc_local = 0;			/* not local path **/
    } 
 }
 u_int32_t lnum;
 lnum = 0;
 for(argn = 1; argn < argc; ++argn) {
    if(strcmp(args[argn], "--debug") == 0) {
	ustc_debug_mode = 1;		/* debug enable **/
    } else
    if(strcmp(args[argn], "--help") == 0) {
	ustc_usage();			/* print help usage & exit **/
	exit(1);
    } else
    if(strcmp(args[argn], "--license") == 0) {
	ustc_license();
	exit(1);
    } else
    if(strcmp(args[argn], "--license-ru") == 0) {
	ustc_license_ru();
	exit(1);
    } else
    if(strcmp(args[argn], "--version") == 0) {
	ustc_version();			/* print version & exit **/
	exit(1);			
    } else
    if(strcmp(args[argn], "--todo") == 0) {
	ustc_todo();			/* print todo note & exit **/
	exit(1);
    } else
    if(strcmp(args[argn], "--changelog") == 0) {
	ustc_changelog();		/* print change log & exit **/
	exit(1);
    } else
    if(strcmp(args[argn], "--man") == 0) {
	ustc_man();			/* print man page and exit **/
	exit(1);
    } else
    if(
	(strcmp(args[argn], "--sort") == 0)
    ) {
	u_int8_t *enb; 
	if(strcmp(args[argn], "--sort") == 0) {
	    enb = &ustc_sort_enable;
	} else 
	{
	    enb = 0;
	}
	if(argn + 1 >= argc) {		/* argument count check */
	    ustc_error(			/* error message print */
		"Option {%s} requiry argument\n",
		args[argn]
	    );
	    exit(1);			/* exit **/
	} else {
	    if(				/* option on check */
		(strcmp((char *)args[argn+1],"enable") == 0) ||
		(strcmp((char *)args[argn+1],"on") == 0) ||
		(strcmp((char *)args[argn+1],"ON") == 0) 
	    ) {
		if(enb != 0) {
		    *enb = 1;		/* option enable */
		}
	    } else
	    if(				/* option off check */
		(strcmp((char *)args[argn+1],"disable") == 0) ||
		(strcmp((char *)args[argn+1],"off") == 0) ||
		(strcmp((char *)args[argn+1],"OFF") == 0)
	    ) {
		if(enb != 0) {
		    *enb = 0;		/* option disable */
		}
	    } else {
		ustc_error("Option --sort requiry argument on or off\n");
		exit(1);			/* exit **/
	    }
	    argn += 1;
	}
    }
    if(					/* part option check */
	(strcmp(args[argn], "--part") == 0)
    ) {
	if(argn + 1 >= argc) {		/* argument count check */
	    ustc_error(			/* print error message */
		"Option %s requiry argument\r\n",
		args[argn]
	    );
	    exit(1);
	} else {
	    snprintf(			/* part write */
	        (char *)ustc_part, 	/* part buffer */
	        sizeof(ustc_part),	/* part buffer length */
	        "%s",			/* format */
	        args[argn+1]		/* part */
	    );
	    argn += 1;			/* next argument */
	}
    } else
    {
	if(argn + 1 == argc - 1) {
	    snprintf(ustc_ifile, sizeof(ustc_ifile), "%s", args[argn]);
	    snprintf(ustc_ofile, sizeof(ustc_ofile), "%s", args[argn+1]);
	    argn += 1;
	} else
	if(argn == argc - 1) {
	    snprintf(ustc_ifile, sizeof(ustc_ifile), "%s", args[argn]);
	}
    }
 }
 ustc_layer_count = lnum;		/* layer number **/
 ustc_printf("\n");
 if(strlen(ustc_ifile) > 0) {
    ustc_fi = fopen(ustc_ifile, "r+b");
    if(ustc_fi == 0) {
	ustc_error("Input file [%s] open error.\n", ustc_ifile);
	return(0);
    }
 }
 if(strlen(ustc_ofile) > 0) {
    ustc_fo = fopen(ustc_ofile, "w+b");
    if(ustc_fo == 0) {
	ustc_error("Input file [%s] open error.\n", ustc_ofile);
	return(0);
    }
 }
 if(1) {
    if((ustc_fi != 0)) {
	ustc_ifmt = ustc_format(ustc_ifile);	/* input file format **/
	ustc_ofmt = ustc_format(ustc_ofile);	/* output file format **/
	if(ustc_fo == 0) {
	    ustc_fo = stdout;
	    if(ustc_ifmt == USTC_LIB) {
		if(ustc_lib_recover() == 0) {
		    rvl = 0;
		}
	    } else
	    if(ustc_ifmt == USTC_PADS_NET) {
 		if(ustc_padsnet_recover() == 0) {
		    rvl = 0;
		}
	    }
	}
	else {
	    if(
		(ustc_ifmt == USTC_LIB) &&
		(ustc_ofmt == USTC_LIB)
	    ) {
		if(ustc_part[0] != '\0') {
			if(ustc_lib_part() == 0) {
			    rvl = 0;
			}
		} else {
		    if(ustc_lib_recover() == 0) {
			rvl = 0;
		    }
		}
	    } else
	    if(
		(ustc_ifmt == USTC_PADS_NET) &&
		(ustc_ofmt == USTC_PADS_NET)
	    ) {
		if(ustc_padsnet_recover() == 0) {
		    rvl = 0;
		}
	    }
	}
    }
 }
 if(ustc_fi != 0) {
    fclose(ustc_fi);
    ustc_fi = 0;
 } 
 if(ustc_fo != 0) {
    fclose(ustc_fo);
    ustc_fo = 0;
 } 
 if(ustc_ibuf != 0) {			/* input buffer **/
    free(ustc_ibuf);			/* free input buffer **/
    ustc_ibuf = 0;			/* clear input buffer pointer **/
    ustc_isize = 0;			/* input buffer size **/
 }    
 if(ustc_obuf != 0) {			/* output buffer **/
    free(ustc_obuf);			/* free output buffer **/
    ustc_obuf = 0;			/* clear output buffer pointer **/
    ustc_osize = 0;			/* output buffer size **/
 }    
 return(rvl);
}
